@@ -39,6 +39,7 @@
 #include <linux/if_ether.h>
 #include <linux/filter.h>
 #include <net/if_arp.h>
+#include <errno.h>
 
 #include "io.h"
 #include "util.h"
@@ -78,9 +79,11 @@ uint16_t _dhcp_checksumv(const struct iovec *iov, size_t iov_cnt)
 			sum += check[i];
 	}
 
+	j--;
+
 	if (len & 0x01) {
 		const uint8_t *odd = iov[j].iov_base;
-		sum += odd[len - 1];
+		sum += odd[iov[j].iov_len - 1];
 	}
 
 	while (sum >> 16)
@@ -166,8 +169,8 @@ static void dhcp_set_ip_udp_headers(struct iphdr *ip, struct udphdr *udp,
 	ip->tot_len = L_CPU_TO_BE16(len + sizeof(*ip) + sizeof(*udp));
 
 	ip->protocol = IPPROTO_UDP;
-	ip->saddr = L_CPU_TO_BE32(saddr);
-	ip->daddr = L_CPU_TO_BE32(daddr);
+	ip->saddr = saddr;
+	ip->daddr = daddr;
 
 	udp->source = L_CPU_TO_BE16(sport);
 	udp->dest = L_CPU_TO_BE16(dport);
@@ -192,9 +195,10 @@ static void dhcp_set_ip_udp_headers(struct iphdr *ip, struct udphdr *udp,
 	ip->check = _dhcp_checksumv(iov, 1);
 }
 
-static int _dhcp_default_transport_broadcast(struct dhcp_transport *s,
+static int _dhcp_default_transport_l2_send(struct dhcp_transport *s,
 						uint32_t saddr, uint16_t sport,
 						uint32_t daddr, uint16_t dport,
+						const uint8_t *dest_mac,
 						const void *data, size_t len)
 {
 	struct dhcp_default_transport *transport =
@@ -219,7 +223,10 @@ static int _dhcp_default_transport_broadcast(struct dhcp_transport *s,
 	addr.sll_protocol = htons(ETH_P_IP);
 	addr.sll_ifindex = s->ifindex;
 	addr.sll_halen = ETH_ALEN;
-	memset(addr.sll_addr, 0xff, ETH_ALEN);
+	if (!dest_mac)
+		memset(addr.sll_addr, 0xff, ETH_ALEN);
+	else
+		memcpy(addr.sll_addr, dest_mac, ETH_ALEN);
 
 	memset(&msg, 0, sizeof(msg));
 	msg.msg_name = &addr;
@@ -269,7 +276,7 @@ static int kernel_udp_socket_open(const char *ifname,
 		.filter = filter
 	};
 
-	s = socket(AF_INET, SOCK_DGRAM, 0);
+	s = socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (s < 0)
 		return -errno;
 
@@ -380,9 +387,11 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 		/* A <- DHCP op */
 		BPF_STMT(BPF_LD + BPF_B + BPF_ABS,
 				offsetof(struct dhcp_packet, dhcp.op)),
-		/* op == BOOTREPLY ? */
+		/* op == (BOOTREPLY | BOOTREQUEST) ? */
 		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K,
-				DHCP_OP_CODE_BOOTREPLY, 1, 0),
+				(port == DHCP_PORT_CLIENT) ?
+				DHCP_OP_CODE_BOOTREPLY :
+				DHCP_OP_CODE_BOOTREQUEST, 1, 0),
 		/* ignore */
 		BPF_STMT(BPF_RET + BPF_K, 0),
 		/* A <- DHCP header type */
@@ -392,6 +401,10 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ARPHRD_ETHER, 1, 0),
 		/* ignore */
 		BPF_STMT(BPF_RET + BPF_K, 0),
+		/* A <- UDP destination port */
+		BPF_STMT(BPF_LD + BPF_H + BPF_ABS,
+				offsetof(struct dhcp_packet, udp.dest)),
+		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, DHCP_PORT_SERVER, 3, 0),
 		/* A <- client identifier */
 		BPF_STMT(BPF_LD + BPF_W + BPF_ABS,
 				offsetof(struct dhcp_packet, dhcp.xid)),
@@ -493,7 +506,7 @@ struct dhcp_transport *_dhcp_default_transport_new(uint32_t ifindex,
 	transport->super.bind = _dhcp_default_transport_bind;
 	transport->super.close = _dhcp_default_transport_close;
 	transport->super.send = _dhcp_default_transport_send;
-	transport->super.broadcast = _dhcp_default_transport_broadcast;
+	transport->super.l2_send = _dhcp_default_transport_l2_send;
 
 	transport->super.ifindex = ifindex;
 	l_strlcpy(transport->ifname, ifname, IFNAMSIZ);
