@@ -55,6 +55,7 @@ void network_set_info(struct network *network, struct network_info *info);
 int network_autoconnect(struct network *network, struct scan_bss *bss);
 void network_connect_failed(struct network *network, bool in_handshake);
 bool network_bss_add(struct network *network, struct scan_bss *bss);
+bool network_bss_update(struct network *network, struct scan_bss *bss);
 bool network_bss_list_isempty(struct network *network);
 void network_bss_list_clear(struct network *network);
 struct scan_bss *network_bss_list_pop(struct network *network);
@@ -78,3 +79,8 @@ void network_blacklist_add(struct network *network, struct scan_bss *bss);
 
 const struct iovec *network_get_extra_ies(struct network *network,
 						size_t *num_elems);
+
+bool network_has_erp_identity(struct network *network);
+
+const struct l_queue_entry *network_bss_list_get_entries(
+						struct network *network);
