@@ -35,6 +35,7 @@
 #include <linux/if_ether.h>
 #include <ell/ell.h>
 
+#include "ell/useful.h"
 #include "linux/nl80211.h"
 #include "src/iwd.h"
 #include "src/module.h"
@@ -50,8 +51,10 @@
 #include "src/mpdu.h"
 #include "src/scan.h"
 
-#define SCAN_MAX_INTERVAL 320
-#define SCAN_INIT_INTERVAL 10
+/* User configurable options */
+static double RANK_5G_FACTOR;
+static uint32_t SCAN_MAX_INTERVAL;
+static uint32_t SCAN_INIT_INTERVAL;
 
 static struct l_queue *scan_contexts;
 
@@ -97,6 +100,11 @@ struct scan_context {
 	unsigned int start_cmd_id;
 	/* Non-zero if GET_SCAN is still running */
 	unsigned int get_scan_cmd_id;
+	/*
+	 * Special request used for getting scan results after the firmware
+	 * roamed automatically.
+	 */
+	unsigned int get_fw_scan_cmd_id;
 	/*
 	 * Whether the top request in the queue has triggered the current
 	 * scan.  May be set and cleared multiple times during a single
@@ -160,7 +168,7 @@ static void scan_request_failed(struct scan_context *sc,
 	if (sr->trigger)
 		sr->trigger(err, sr->userdata);
 	else if (sr->callback)
-		sr->callback(err, NULL, sr->userdata);
+		sr->callback(err, NULL, NULL, sr->userdata);
 
 	wiphy_radio_work_done(sc->wiphy, sr->work.id);
 }
@@ -205,6 +213,9 @@ static void scan_context_free(struct scan_context *sc)
 	if (sc->get_scan_cmd_id && nl80211)
 		l_genl_family_cancel(nl80211, sc->get_scan_cmd_id);
 
+	if (sc->get_fw_scan_cmd_id && nl80211)
+		l_genl_family_cancel(nl80211, sc->get_fw_scan_cmd_id);
+
 	l_free(sc);
 }
 
@@ -297,7 +308,7 @@ static void scan_build_attr_ie(struct l_genl_msg *msg,
 	iov[iov_elems].iov_len = ext_capa[1] + 2;
 	iov_elems++;
 
-	if (util_is_bit_set(ext_capa[2 + 3], 7)) {
+	if (test_bit(&ext_capa[2 + 3], 7)) {
 		/* Order 12 - Interworking */
 		interworking[0] = IE_TYPE_INTERWORKING;
 		interworking[1] = 1;
@@ -361,6 +372,18 @@ static struct l_genl_msg *scan_build_cmd(struct scan_context *sc,
 		 */
 		flags |= NL80211_SCAN_FLAG_RANDOM_ADDR;
 
+	if (!is_passive && params->source_mac &&
+			wiphy_can_randomize_mac_addr(sc->wiphy)) {
+		static const uint8_t mask[6] =	/* No random bits */
+			{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
+
+		flags |= NL80211_SCAN_FLAG_RANDOM_ADDR;
+		l_genl_msg_append_attr(msg, NL80211_ATTR_MAC, 6,
+					params->source_mac);
+		l_genl_msg_append_attr(msg, NL80211_ATTR_MAC_MASK, 6,
+					mask);
+	}
+
 	if (!is_passive && wiphy_has_ext_feature(sc->wiphy,
 					NL80211_EXT_FEATURE_SCAN_RANDOM_SN))
 		flags |= NL80211_SCAN_FLAG_RANDOM_SN;
@@ -424,6 +447,35 @@ done:
 	return msg;
 }
 
+struct l_genl_msg *scan_build_trigger_scan_bss(uint32_t ifindex,
+						struct wiphy *wiphy,
+						uint32_t frequency,
+						const uint8_t *ssid,
+						uint32_t ssid_len)
+{
+	struct l_genl_msg *msg = l_genl_msg_new(NL80211_CMD_TRIGGER_SCAN);
+	uint32_t flags = 0;
+
+	l_genl_msg_append_attr(msg, NL80211_ATTR_IFINDEX, 4, &ifindex);
+
+	l_genl_msg_enter_nested(msg, NL80211_ATTR_SCAN_FREQUENCIES);
+	l_genl_msg_append_attr(msg, 0, 4, &frequency);
+	l_genl_msg_leave_nested(msg);
+
+	if (wiphy_has_ext_feature(wiphy, NL80211_EXT_FEATURE_SCAN_RANDOM_SN))
+		flags |= NL80211_SCAN_FLAG_RANDOM_SN;
+
+	if (flags)
+		l_genl_msg_append_attr(msg, NL80211_ATTR_SCAN_FLAGS, 4, &flags);
+
+	/* direct probe request scan */
+	l_genl_msg_enter_nested(msg, NL80211_ATTR_SCAN_SSIDS);
+	l_genl_msg_append_attr(msg, 0, ssid_len, ssid);
+	l_genl_msg_leave_nested(msg);
+
+	return msg;
+}
+
 struct scan_cmds_add_data {
 	struct scan_context *sc;
 	const struct scan_parameters *params;
@@ -646,6 +698,7 @@ bool scan_cancel(uint64_t wdev_id, uint32_t id)
 		sc->start_cmd_id = 0;
 		l_queue_remove(sc->requests, sr);
 		sc->started = false;
+		sc->work_started = false;
 	} else
 		l_queue_remove(sc->requests, sr);
 
@@ -670,6 +723,7 @@ static void scan_periodic_triggered(int err, void *user_data)
 }
 
 static bool scan_periodic_notify(int err, struct l_queue *bss_list,
+					const struct scan_freq_set *freqs,
 					void *user_data)
 {
 	struct scan_context *sc = user_data;
@@ -677,7 +731,7 @@ static bool scan_periodic_notify(int err, struct l_queue *bss_list,
 	scan_periodic_rearm(sc);
 
 	if (sc->sp.callback)
-		return sc->sp.callback(err, bss_list, sc->sp.userdata);
+		return sc->sp.callback(err, bss_list, freqs, sc->sp.userdata);
 
 	return false;
 }
@@ -806,6 +860,8 @@ static void scan_periodic_timeout(struct l_timeout *timeout, void *user_data)
 	l_debug("scan_periodic_timeout: %" PRIx64, sc->wdev_id);
 
 	sc->sp.interval *= 2;
+	if (sc->sp.interval > SCAN_MAX_INTERVAL)
+		sc->sp.interval = SCAN_MAX_INTERVAL;
 
 	scan_periodic_queue(sc);
 }
@@ -943,18 +999,6 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 			bss->ssid_len = iter.len;
 			have_ssid = true;
 			break;
-		case IE_TYPE_SUPPORTED_RATES:
-			if (iter.len > 8)
-				return false;
-
-			bss->has_sup_rates =  true;
-			memcpy(bss->supp_rates_ie, iter.data - 2, iter.len + 2);
-
-			break;
-		case IE_TYPE_EXTENDED_SUPPORTED_RATES:
-			bss->ext_supp_rates_ie = l_memdup(iter.data - 2,
-								iter.len + 2);
-			break;
 		case IE_TYPE_RSN:
 			if (!bss->rsne)
 				bss->rsne = l_memdup(iter.data - 2,
@@ -1000,20 +1044,10 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 
 			break;
 		case IE_TYPE_HT_CAPABILITIES:
-			if (iter.len != 26)
-				return false;
-
 			bss->ht_capable = true;
-			memcpy(bss->ht_ie, iter.data - 2, iter.len + 2);
-
 			break;
 		case IE_TYPE_VHT_CAPABILITIES:
-			if (iter.len != 12)
-				return false;
-
 			bss->vht_capable = true;
-			memcpy(bss->vht_ie, iter.data - 2, iter.len + 2);
-
 			break;
 		case IE_TYPE_ADVERTISEMENT_PROTOCOL:
 			if (iter.len < 2)
@@ -1107,7 +1141,9 @@ static bool scan_parse_bss_information_elements(struct scan_bss *bss,
 	return have_ssid;
 }
 
-static struct scan_bss *scan_parse_attr_bss(struct l_genl_attr *attr)
+static struct scan_bss *scan_parse_attr_bss(struct l_genl_attr *attr,
+						struct wiphy *wiphy,
+						uint32_t *out_seen_ms_ago)
 {
 	uint16_t type, len;
 	const void *data;
@@ -1164,6 +1200,18 @@ static struct scan_bss *scan_parse_attr_bss(struct l_genl_attr *attr)
 			beacon_ies = data;
 			beacon_ies_len = len;
 			break;
+		case NL80211_BSS_SEEN_MS_AGO:
+			if (L_WARN_ON(len != sizeof(uint32_t)))
+				break;
+
+			*out_seen_ms_ago = l_get_u32(data);
+			break;
+		case NL80211_BSS_LAST_SEEN_BOOTTIME:
+			if (L_WARN_ON(len != sizeof(uint64_t)))
+				break;
+
+			bss->time_stamp = l_get_u64(data) / L_NSEC_PER_USEC;
+			break;
 		}
 	}
 
@@ -1178,8 +1226,16 @@ static struct scan_bss *scan_parse_attr_bss(struct l_genl_attr *attr)
 				memcmp(ies, beacon_ies, ies_len)))
 		bss->source_frame = SCAN_BSS_PROBE_RESP;
 
-	if (ies && !scan_parse_bss_information_elements(bss, ies, ies_len))
-		goto fail;
+	/* Set data rate to something low, just in case estimation fails */
+	bss->data_rate = 2000000;
+
+	if (ies) {
+		if (!scan_parse_bss_information_elements(bss, ies, ies_len))
+			goto fail;
+
+		L_WARN_ON(wiphy_estimate_data_rate(wiphy, ies, ies_len, bss,
+						&bss->data_rate) < 0);
+	}
 
 	return bss;
 
@@ -1211,87 +1267,43 @@ static struct scan_freq_set *scan_parse_attr_scan_frequencies(
 }
 
 static struct scan_bss *scan_parse_result(struct l_genl_msg *msg,
-						uint64_t *out_wdev)
+						struct wiphy *wiphy,
+						uint32_t *out_seen_ms_ago)
 {
 	struct l_genl_attr attr, nested;
-	uint16_t type, len;
-	const void *data;
-	const uint64_t *wdev = NULL;
+	uint16_t type;
 	struct scan_bss *bss = NULL;
 
 	if (!l_genl_attr_init(&attr, msg))
 		return NULL;
 
-	while (l_genl_attr_next(&attr, &type, &len, &data)) {
+	while (l_genl_attr_next(&attr, &type, NULL, NULL)) {
 		switch (type) {
-		case NL80211_ATTR_WDEV:
-			if (len != sizeof(uint64_t))
-				return NULL;
-
-			wdev = data;
-			break;
-
 		case NL80211_ATTR_BSS:
 			if (!l_genl_attr_recurse(&attr, &nested))
 				return NULL;
 
-			bss = scan_parse_attr_bss(&nested);
+			bss = scan_parse_attr_bss(&nested, wiphy,
+							out_seen_ms_ago);
 			break;
 		}
 	}
 
-	if (!bss)
-		return NULL;
-
-	if (!wdev) {
-		scan_bss_free(bss);
-		return NULL;
-	}
-
-	if (out_wdev)
-		*out_wdev = *wdev;
-
 	return bss;
 }
 
-/* User configurable options */
-static double RANK_5G_FACTOR;
-
 static void scan_bss_compute_rank(struct scan_bss *bss)
 {
-	static const double RANK_RSNE_FACTOR = 1.2;
-	static const double RANK_WPA_FACTOR = 1.0;
-	static const double RANK_OPEN_FACTOR = 0.5;
-	static const double RANK_NO_PRIVACY_FACTOR = 0.5;
 	static const double RANK_HIGH_UTILIZATION_FACTOR = 0.8;
 	static const double RANK_LOW_UTILIZATION_FACTOR = 1.2;
-	static const double RANK_MIN_SUPPORTED_RATE_FACTOR = 0.6;
-	static const double RANK_MAX_SUPPORTED_RATE_FACTOR = 1.3;
 	double rank;
 	uint32_t irank;
-
 	/*
-	 * Signal strength is in mBm (100 * dBm) and is negative.
-	 * WiFi range is -0 to -100 dBm
+	 * Maximum rate is 2340Mbps (VHT)
 	 */
+	double max_rate = 2340000000;
 
-	/* Heavily slanted towards signal strength */
-	rank = 10000 + bss->signal_strength;
-
-	/*
-	 * Prefer RSNE first, WPA second.  Open networks are much less
-	 * desirable.
-	 */
-	if (bss->rsne)
-		rank *= RANK_RSNE_FACTOR;
-	else if (bss->wpa)
-		rank *= RANK_WPA_FACTOR;
-	else
-		rank *= RANK_OPEN_FACTOR;
-
-	/* We prefer networks with CAP PRIVACY */
-	if (!(bss->capability & IE_BSS_CAP_PRIVACY))
-		rank *= RANK_NO_PRIVACY_FACTOR;
+	rank = (double)bss->data_rate / max_rate * USHRT_MAX;
 
 	/* Prefer 5G networks over 2.4G */
 	if (bss->frequency > 4000)
@@ -1303,29 +1315,6 @@ static void scan_bss_compute_rank(struct scan_bss *bss)
 	else if (bss->utilization <= 63)
 		rank *= RANK_LOW_UTILIZATION_FACTOR;
 
-	if (bss->has_sup_rates || bss->ext_supp_rates_ie) {
-		uint64_t data_rate;
-
-		if (ie_parse_data_rates(bss->has_sup_rates ?
-					bss->supp_rates_ie : NULL,
-					bss->ext_supp_rates_ie,
-					bss->ht_capable ? bss->ht_ie : NULL,
-					bss->vht_capable ? bss->vht_ie : NULL,
-					bss->signal_strength / 100,
-					&data_rate) == 0) {
-			double factor = RANK_MAX_SUPPORTED_RATE_FACTOR -
-					RANK_MIN_SUPPORTED_RATE_FACTOR;
-
-			/*
-			 * Maximum rate is 2340Mbps (VHT)
-			 */
-			factor = factor * data_rate / 2340000000U +
-						RANK_MIN_SUPPORTED_RATE_FACTOR;
-			rank *= factor;
-		} else
-			rank *= RANK_MIN_SUPPORTED_RATE_FACTOR;
-	}
-
 	irank = rank;
 
 	if (irank > USHRT_MAX)
@@ -1352,7 +1341,6 @@ struct scan_bss *scan_bss_new_from_probe_req(const struct mmpdu_header *mpdu,
 	if (!scan_parse_bss_information_elements(bss, body, body_len))
 		goto fail;
 
-	scan_bss_compute_rank(bss);
 	return bss;
 
 fail:
@@ -1362,7 +1350,6 @@ fail:
 
 void scan_bss_free(struct scan_bss *bss)
 {
-	l_free(bss->ext_supp_rates_ie);
 	l_free(bss->rsne);
 	l_free(bss->wpa);
 	l_free(bss->wsc);
@@ -1446,20 +1433,26 @@ static void get_scan_callback(struct l_genl_msg *msg, void *user_data)
 	struct scan_context *sc = results->sc;
 	struct scan_bss *bss;
 	uint64_t wdev_id;
+	uint32_t seen_ms_ago = 0;
 
 	l_debug("get_scan_callback");
 
-	bss = scan_parse_result(msg, &wdev_id);
-	if (!bss)
+	if (nl80211_parse_attrs(msg, NL80211_ATTR_WDEV, &wdev_id,
+					NL80211_ATTR_UNSPEC) < 0)
 		return;
 
 	if (wdev_id != sc->wdev_id) {
 		l_warn("wdev mismatch in get_scan_callback");
-		scan_bss_free(bss);
 		return;
 	}
 
-	bss->time_stamp = results->time_stamp;
+	bss = scan_parse_result(msg, sc->wiphy, &seen_ms_ago);
+	if (!bss)
+		return;
+
+	if (!bss->time_stamp)
+		bss->time_stamp = results->time_stamp -
+					seen_ms_ago * L_USEC_PER_MSEC;
 
 	scan_bss_compute_rank(bss);
 	l_queue_insert(results->bss_list, bss, scan_bss_rank_compare, NULL);
@@ -1483,6 +1476,7 @@ static void discover_hidden_network_bsses(struct scan_context *sc,
 
 static void scan_finished(struct scan_context *sc,
 				int err, struct l_queue *bss_list,
+				const struct scan_freq_set *freqs,
 				struct scan_request *sr)
 {
 	bool new_owner = false;
@@ -1496,7 +1490,8 @@ static void scan_finished(struct scan_context *sc,
 		sc->work_started = false;
 
 		if (sr->callback)
-			new_owner = sr->callback(err, bss_list, sr->userdata);
+			new_owner = sr->callback(err, bss_list,
+							freqs, sr->userdata);
 
 		/*
 		 * Can start a new scan now that we've removed this one from
@@ -1507,7 +1502,8 @@ static void scan_finished(struct scan_context *sc,
 		 */
 		wiphy_radio_work_done(sc->wiphy, sr->work.id);
 	} else if (sc->sp.callback)
-		new_owner = sc->sp.callback(err, bss_list, sc->sp.userdata);
+		new_owner = sc->sp.callback(err, bss_list,
+						freqs, sc->sp.userdata);
 
 	if (bss_list && !new_owner)
 		l_queue_destroy(bss_list,
@@ -1524,7 +1520,8 @@ static void get_scan_done(void *user)
 	sc->get_scan_cmd_id = 0;
 
 	if (l_queue_peek_head(sc->requests) == results->sr)
-		scan_finished(sc, 0, results->bss_list, results->sr);
+		scan_finished(sc, 0, results->bss_list,
+						results->freqs, results->sr);
 	else
 		l_queue_destroy(results->bss_list,
 				(l_queue_destroy_func_t) scan_bss_free);
@@ -1639,7 +1636,7 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 			sc->triggered = false;
 
 			if (!sr->callback) {
-				scan_finished(sc, -ECANCELED, NULL, sr);
+				scan_finished(sc, -ECANCELED, NULL, NULL, sr);
 				break;
 			}
 
@@ -1662,7 +1659,7 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 
 			/* An external scan may have flushed our results */
 			if (sc->started && scan_parse_flush_flag_from_msg(msg))
-				scan_finished(sc, -EAGAIN, NULL, sr);
+				scan_finished(sc, -EAGAIN, NULL, NULL, sr);
 			else
 				send_next = true;
 
@@ -1726,7 +1723,7 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 		if (sc->triggered) {
 			sc->triggered = false;
 
-			scan_finished(sc, -ECANCELED, NULL, sr);
+			scan_finished(sc, -ECANCELED, NULL, NULL, sr);
 		} else {
 			/*
 			 * If this was an external scan that got aborted
@@ -1743,6 +1740,72 @@ static void scan_notify(struct l_genl_msg *msg, void *user_data)
 	}
 }
 
+static void get_fw_scan_done(void *userdata)
+{
+	struct scan_results *results = userdata;
+	struct scan_request *sr = results->sr;
+	struct scan_context *sc = results->sc;
+	int err = l_queue_length(results->bss_list) == 0 ? -ENOENT : 0;
+	bool new_owner = false;
+
+	sc->get_fw_scan_cmd_id = 0;
+
+	if (sr->callback)
+		new_owner = sr->callback(err, results->bss_list, NULL,
+						sr->userdata);
+
+	if (!new_owner)
+		l_queue_destroy(results->bss_list,
+				(l_queue_destroy_func_t) scan_bss_free);
+
+	if (sr->destroy)
+		sr->destroy(sr->userdata);
+
+	l_free(sr);
+	l_free(results);
+}
+
+bool scan_get_firmware_scan(uint64_t wdev_id, scan_notify_func_t notify,
+				void *userdata, scan_destroy_func_t destroy)
+{
+	struct l_genl_msg *scan_msg;
+	struct scan_results *results;
+	struct scan_request *sr;
+	struct scan_context *sc = l_queue_find(scan_contexts,
+						scan_context_match, &wdev_id);
+
+	if (!sc)
+		return false;
+
+	sr = l_new(struct scan_request, 1);
+	sr->callback = notify;
+	sr->destroy = destroy;
+	sr->userdata = userdata;
+
+	results = l_new(struct scan_results, 1);
+	results->sc = sc;
+	results->time_stamp = l_time_now();
+	results->bss_list = l_queue_new();
+	results->sr = sr;
+
+	scan_msg = l_genl_msg_new_sized(NL80211_CMD_GET_SCAN, 8);
+	l_genl_msg_append_attr(scan_msg, NL80211_ATTR_WDEV, 8, &sc->wdev_id);
+
+	sc->get_fw_scan_cmd_id = l_genl_family_dump(nl80211, scan_msg,
+							get_scan_callback,
+							results,
+							get_fw_scan_done);
+	if (!sc->get_fw_scan_cmd_id) {
+		l_queue_destroy(results->bss_list,
+				(l_queue_destroy_func_t) scan_bss_free);
+		l_free(results);
+		l_free(sr);
+		return false;
+	}
+
+	return true;
+}
+
 uint8_t scan_freq_to_channel(uint32_t freq, enum scan_band *out_band)
 {
 	uint32_t channel = 0;
@@ -1977,7 +2040,7 @@ bool scan_freq_set_add(struct scan_freq_set *freqs, uint32_t freq)
 	return false;
 }
 
-bool scan_freq_set_contains(struct scan_freq_set *freqs, uint32_t freq)
+bool scan_freq_set_contains(const struct scan_freq_set *freqs, uint32_t freq)
 {
 	enum scan_band band;
 	uint8_t channel;
@@ -2149,6 +2212,20 @@ static int scan_init(void)
 					&RANK_5G_FACTOR))
 		RANK_5G_FACTOR = 1.0;
 
+	if (!l_settings_get_uint(config, "Scan", "InitialPeriodicScanInterval",
+					&SCAN_INIT_INTERVAL))
+		SCAN_INIT_INTERVAL = 10;
+
+	if (SCAN_INIT_INTERVAL > UINT16_MAX)
+		SCAN_INIT_INTERVAL = UINT16_MAX;
+
+	if (!l_settings_get_uint(config, "Scan", "MaximumPeriodicScanInterval",
+					&SCAN_MAX_INTERVAL))
+		SCAN_MAX_INTERVAL = 300;
+
+	if (SCAN_MAX_INTERVAL > UINT16_MAX)
+		SCAN_MAX_INTERVAL = UINT16_MAX;
+
 	return 0;
 }
 
